"""
Exceptions.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

_exception_notes = hasattr(BaseException, 'add_note')

__all__ = [
    'add_note',
    'EdzedError', 'EdzedCircuitError', 'EdzedInvalidState', 'EdzedUnknownEvent']

class EdzedError(Exception):
    """Base class for Edzed exceptions."""

class EdzedCircuitError(EdzedError):
    """Critical error."""

class EdzedInvalidState(EdzedError):
    """Invalid state error."""

class EdzedUnknownEvent(EdzedError):
    """Event type not supported."""

def add_note(exc: BaseException, note:str) -> None:
    """Add a note to an exception."""
    if _exception_notes:
        # supported natively in Python >= 3.11
        exc.add_note(note)
    elif exc.args and isinstance(exc.args[0], str):
        # fallback: prepend the note to the error message
        exc.args = (f"[{note}] {exc.args[0]}", *exc.args[1:])



"""
Sequential blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from __future__ import annotations

import asyncio
from collections.abc import (
    Callable, Collection, Coroutine, Iterator, Mapping, MutableSet, Sequence)
import concurrent.futures
import itertools
import functools
from typing import Any, Optional, Type
import weakref

from .. import addons
from .. import block
from .. import fsm
from .. import utils


__all__ = ['InitAsync', 'Input', 'InputExp', 'InExecutor', 'OutputAsync', 'OutputFunc']


class _Validation:
    """Value validation mix-in."""

    def __init__(
            self, *args,
            schema: Optional[Callable[[Any], Any]] = None,
            check: Optional[Callable[[Any], Any]] = None,
            allowed: Optional[Collection] = None,
            **kwargs):
        self._schema = schema
        self._check = check
        self._allowed = None if allowed is None else frozenset(allowed)
        super().__init__(*args, **kwargs)

    def _validate(self, value: Any) -> Any:
        """
        Validate a value.

        Return the result if the value is accepted.
        Raise a ValueError if not.
        """
        if self._allowed is not None and value not in self._allowed:
            raise ValueError(f"Validation error: {value!r} is not among allowed values")
        if self._check is not None and not self._check(value):
            raise ValueError(f"Validation function rejected value {value!r}")
        if self._schema is not None:
            try:
                value = self._schema(value)
            except Exception as err:
                raise ValueError(
                    f"Validation schema rejected value {value!r} with error: {err}") from None
        return value


class Input(_Validation, addons.AddonPersistence, block.SBlock):
    """
    Input with optional value validation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.initdef is not block.UNDEF:
            self._validate(self.initdef)

    def init_from_value(self, value: Any) -> None:
        self.put(value)

    def _event_put(self, *, value: Any, **_data) -> bool:
        try:
            value = self._validate(value)
        except ValueError as err:
            self.log_warning("%s", err)
            return False
        self.set_output(value)
        return True

    _restore_state = init_from_value


class InputExp(_Validation, fsm.FSM):
    """
    Input with an expiration time.
    """

    STATES = ['expired', 'valid']
    TIMERS = {
        'valid': (None, fsm.Goto('expired')),
        }
    EVENTS = (
        ('put', None, 'valid'),
        )

    def __init__(
            self, *args,
            duration: Optional[int|float|str],
            expired: Any = None,
            initdef: Any = block.UNDEF,
            **kwargs):
        # meaning of the initdef parameter:
        # - in InputExp: initial value
        # - in the underlying FSM: initial state
        has_init_value = initdef is not block.UNDEF
        super().__init__(
            *args,
            t_valid=duration,
            initdef = 'valid' if has_init_value else 'expired',
            **kwargs)
        if has_init_value:
            self.sdata['input'] = self._validate(initdef)
        self._expired = self._validate(expired)

    def cond_put(self) -> bool:
        data = fsm.fsm_event_data.get()
        value = data['value']
        try:
            value = self._validate(value)
        except ValueError as err:
            self.log_warning("%s", err)
            return False
        self.sdata['input'] = value
        return True

    def on_enter_expired(self) -> None:
        self.sdata.pop('input', None)

    def calc_output(self) -> Any:
        """Stop the FSM part from setting the output."""
        return self.sdata['input'] if self._state == 'valid' else self._expired


def _check_arg(name, arg: Any) -> None:
    """
    Verify that arg is a sequence of strings.

    An early check should prevent difficult to understand error messages.
    """
    if isinstance(arg, str) or not isinstance(arg, Sequence) \
            or any(not isinstance(k, str) for k in arg):
        raise TypeError(
            f"Argument {name!r} should be a sequence (e.g. a list or tuple) of strings, "
            f"but got {arg!r}")


def _args_as_string(args: Sequence, kwargs: Mapping) -> str:
    """Convert args and kwargs to a printable string."""
    return '(' + ', '.join(itertools.chain(
        (repr(a) for a in args),
        (f"{k}={v!r}" for k, v in kwargs.items()))) + ')'


class OutputAsync(addons.AddonAsync, block.SBlock):
    """
    Run a coroutine as an output task when a value arrives.
    """

    def __init__(
            self, *args,
            coro: Callable[..., Coroutine], # i.e. an async function
            mode: str,
            f_args: Sequence[str] = ('value',),
            f_kwargs: Sequence[str] = (),
            guard_time: Optional[int|float|str] = None,
            on_success:
                Optional[block.Event|Iterator[block.Event]|Sequence[block.Event]] = None,
            on_cancel: Optional[block.Event|Iterator[block.Event]|Sequence[block.Event]] = None,
            on_error: None|block.Event|Iterator[block.Event]|Sequence[block.Event],
            stop_data: Optional[Mapping[str, Any]] = None,
            **kwargs):
        _check_arg('f_args', f_args)
        _check_arg('f_kwargs', f_args)
        self._on_success = block.event_tuple(on_success)
        self._on_cancel = block.event_tuple(on_cancel)
        self._on_error = block.event_tuple(on_error)
        self._guard_time = 0.0 if guard_time is None else utils.time_period(guard_time)
        self._coro = coro
        if mode in {"c", "cancel"}:
            self._ctrl_coro = self._ctrl_cancel
        elif mode in {"w", "wait"}:
            self._ctrl_coro = self._ctrl_wait
        elif mode in {"s", "start"}:
            self._ctrl_coro = self._ctrl_start
            if self._guard_time > 0.0:
                self.log_warning("using guard_time in 'start' mode is ineffective!")
        else:
            raise ValueError(
                "Argument 'mode' must be one of: 'cancel', 'start', 'wait' "
                f"(may be abbreviated to 'c', 's', or 'w'), but got {mode!r}")
        self._f_args = f_args
        self._f_kwargs = f_kwargs
        self._stop_data = stop_data
        self._ctrl_task: asyncio.Task
        self._queue: asyncio.Queue
        super().__init__(*args, **kwargs)
        if self._guard_time > self.stop_timeout:
            raise ValueError(
                f"guard_time {self._guard_time:.3f} must not exceed "
                f"stop_timeout {self.stop_timeout:.3f} seconds.")

    def _event_put(self, **data) -> None:
        self._queue.put_nowait(data)

    async def _output_coro(self, data: Mapping) -> None:
        args = tuple(data[k] for k in self._f_args)
        kwargs = {k: data[k] for k in self._f_kwargs}
        if self.debug:
            self.log_debug("output task started; args: %s", _args_as_string(args, kwargs))
        try:
            retval = await self._coro(*args, **kwargs)
        except asyncio.CancelledError:
            # it is assumed that the coroutine was cancelled by the control task
            self.log_debug("output task cancelled")
            for ev in self._on_cancel:
                ev.send(self, trigger='cancel', put=data)
        except Exception as err:
            self.log_error(
                "output task failed; args: %s; error: %r", _args_as_string(args, kwargs), err)
            for ev in self._on_error:
                ev.send(self, trigger='error', error=err, put=data)
        else:
            self.log_debug("output task returned value %r", retval)
            for ev in self._on_success:
                ev.send(self, trigger='success', value=retval, put=data)
        if self._guard_time > 0.0:
            try:
                await utils.shield_cancel(asyncio.sleep(self._guard_time))
            except asyncio.CancelledError:
                # shield_cancel re-reaises any CancelledError when it stops shielding
                pass

    async def _output_coro_wrapper(self, data: Mapping) -> None:
        """Count the active tasks."""
        self.set_output(self.output + 1)
        try:
            await self._output_coro(data)
        finally:
            self.set_output(self.output - 1)


    async def _ctrl_cancel(self) -> None:
        """
        Start an output task for data from the queue.

        Only one task may be running at a time. New data
        are processed asap, cancelling the current task if necessary.

        Stop serving after receiving the None sentinel value.
        """
        task = None
        stop = False
        queue = self._queue
        while True:
            if not stop:
                data = await queue.get()
                if data is None:
                    stop = True
            if task and not task.done():
                if not stop:
                    task.cancel()
                # do not use try/await task/except here, because the _output_coro
                # catches all exceptions from user-supplied 'coro'
                await task
            if stop:
                break
            while not queue.empty():
                new_data = queue.get_nowait()
                if new_data is None:
                    stop = True
                    break
                self.log_debug("Discarding: %r", data)
                for ev in self._on_cancel:
                    ev.send(self, trigger='cancel', put=data)
                data = new_data
            # we are already running as a monitored task
            task = asyncio.create_task(self._output_coro_wrapper(data))

    async def _ctrl_wait(self) -> None:
        """
        Start an output task for values from the queue.

        Only one task may be running at a time. New values wait
        in the queue.

        Stop serving after receiving the None sentinel value.
        """
        while True:
            data = await self._queue.get()
            if self.debug:
                qsize = self._queue.qsize()
                if qsize > 0:
                    self.log_debug("%d value(s) waiting in output queue", qsize)
            if data is None:
                break
            await self._output_coro_wrapper(data)

    async def _ctrl_start(self) -> None:
        """
        Start an output task for values from the queue asap.

        Stop serving after receiving the None sentinel value.
        """
        tasks: MutableSet[asyncio.Task] = weakref.WeakSet()
        while True:
            data = await self._queue.get()
            if data is None:
                break
            tasks.add(asyncio.create_task(self._output_coro_wrapper(data)))
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

    def start(self) -> None:
        super().start()
        self._queue = asyncio.Queue()
        self._ctrl_task = self._create_monitored_task(
            self._ctrl_coro(), name=f"edzed: control task for block {self.name!r}")

    def init_regular(self) -> None:
        self.set_output(0)

    def stop(self) -> None:
        # do not compare self._ctrl_coro using "is" (descriptors are in play)
        # pylint: disable=comparison-with-callable
        if self._stop_data is not None and self._ctrl_coro != self._ctrl_start:
            # stop_data processing in start mode moved to stop_async, because
            # that mode does not guarantee that stop_data will be processed last
            self._event_put(**self._stop_data)
        self._queue.put_nowait(None)    # stop serving
        super().stop()

    async def stop_async(self) -> None:
        try:
            await self._ctrl_task
        except asyncio.CancelledError:
            pass
        # pylint: disable=comparison-with-callable
        if self._stop_data is not None and self._ctrl_coro == self._ctrl_start:
            await self._output_coro_wrapper(self._stop_data)
        # super().stop_async is an UNBOUND placeholder, pylint complains about missing 'self'
        # pylint: disable=no-value-for-parameter
        await super().stop_async()


class InExecutor:
    """
    Convert a blocking function into a coroutine usable with OutputAsync.

    The function will be executed in an executor.
    """
    def __init__(
            self,
            func: Callable,
            executor: Type[concurrent.futures.Executor] = concurrent.futures.ThreadPoolExecutor
            ):
        self._func = func
        self._executor = executor

    async def __call__(self, *args, **kwargs):
        run_in_executor = asyncio.get_running_loop().run_in_executor
        with self._executor() as pool:
            if kwargs:
                func = functools.partial(self._func, *args, **kwargs)
                return await run_in_executor(pool, func)
            return await run_in_executor(pool, self._func, *args)


class OutputFunc(block.SBlock):
    """
    Run a function when a value arrives.
    """

    def __init__(
            self, *args,
            func: Callable,
            f_args: Sequence[str] = ('value',),
            f_kwargs: Sequence[str] = (),
            on_success:
                Optional[block.Event|Iterator[block.Event]|Sequence[block.Event]] = None,
            on_error: None|block.Event|Iterator[block.Event]|Sequence[block.Event],
            stop_data: Optional[Mapping[str, Any]] = None,
            **kwargs):
        _check_arg('f_args', f_args)
        _check_arg('f_kwargs', f_kwargs)
        self._on_success = block.event_tuple(on_success)
        self._on_error = block.event_tuple(on_error)
        self._func = func
        self._f_args = f_args
        self._f_kwargs = f_kwargs
        self._stop_data = stop_data
        super().__init__(*args, **kwargs)

    def _event_put(self, **data) -> tuple[str, Any]:
        args = tuple(data[k] for k in self._f_args)
        kwargs = {k: data[k] for k in self._f_kwargs}
        try:
            result = self._func(*args, **kwargs)
        except Exception as err:
            self.log_error(
                "output function failed; args: %s; error: %r",
                _args_as_string(args, kwargs), err)
            for ev in self._on_error:
                ev.send(self, trigger='error', error=err)
            return ('error', err)
        else:
            self.log_debug("output function returned: %r", result)
            for ev in self._on_success:
                ev.send(self, trigger='success', value=result)
            return ('result', result)

    def init_regular(self) -> None:
        """Initialize the internal state."""
        self.set_output(False)

    def stop(self) -> None:
        if self._stop_data is not None:
            self._event_put(**self._stop_data)
        super().stop()


class InitAsync(addons.AddonAsync, block.SBlock):
    """
    Run a coroutine once during the circuit initialization.
    """

    def __init__(self, *args, init_coro: Sequence, **kwargs):
        if not isinstance(init_coro, Sequence):
            raise TypeError(
                "Parameter 'init_coro' must be a sequence (list, tuple, ...), "
                f"but got {init_coro!r}")
        if not init_coro:
            raise ValueError("Parameter 'init_coro' cannot be empty.")
        self._init_coro = init_coro
        super().__init__(*args, **kwargs)

    def init_regular(self) -> None:
        if self.is_initialized() or self.initdef is not block.UNDEF:
            return      # is initialized or will be initialized
        # initialize only to prevent a startup failure
        self._output_events = ()    # do not send output events
        self.set_output(None)

    def init_from_value(self, value: Any) -> None:
        self.set_output(value)

    async def init_async(self) -> None:
        coro, *args = self._init_coro
        result = await coro(*args)
        self.set_output(result)



"""
Simple event-driven zero-delay logic/digital circuit simulation.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine, Iterator, Iterable, MutableMapping, Set
import fnmatch
import logging
import operator
import signal
import time
from typing import Any, NoReturn, Optional

from . import addons
from . import block
from .blocklib import cblocks
from .blocklib import sblocks1
from .exceptions import add_note, EdzedCircuitError, EdzedInvalidState


__all__ = ['get_circuit', 'reset_circuit', 'run']

# limit for oscillation/instability detection:
_MAX_EVALS_PER_BLOCK = 3


_logger = logging.getLogger(__package__)
_current_circuit: Optional[Circuit] = None


def get_circuit() -> Circuit:
    """Get the current circuit. Create one if it does not exist."""
    global _current_circuit

    if _current_circuit is None:
        _current_circuit = Circuit()
    return _current_circuit


def reset_circuit() -> None:
    """
    Clear the circuit and create a new one.
    """
    global _current_circuit

    if _current_circuit is None:
        return
    try:
        _current_circuit.abort(EdzedCircuitError('forced circuit reset'))
         # abort is ignored if not running
    except Exception as err:
        # e.g. RuntimeError: Event loop is closed
        _logger.warning("reset_circuit(): %r error ignored", err)
    _current_circuit = Circuit()


class _BlockResolver:
    """
    A helper allowing to reference a block by name.

    Objects with a reference to be resolved must register themselves.

    The simulator initialization routine calls resolve() to replace
    block names by real block objects in all registered objects.
    """

    def __init__(self, resolve_function: Callable[[str], block.Block]):
        self._unresolved: list[tuple[Any, str, type[block.Block]]] = []
        self._resolve_function = resolve_function

    @staticmethod
    def _check_type(obj, attr: str, blk: block.Block, block_type: type[block.Block]) -> None:
        if not isinstance(blk, block_type):
            raise TypeError(
                f"type of {blk!r} stored in {obj!r} as attribute {attr!r} "
                f"should be {block_type.__name__}")


    def register(self, obj, attr: str, block_type: type[block.Block] = block.Block) -> None:
        """Register an object with a block reference to be resolved."""
        blk = getattr(obj, attr)
        if isinstance(blk, str):
            # name to be resolved
            self._unresolved.append((obj, attr, block_type))
        else:
            # no need to resolve, just check
            self._check_type(obj, attr, blk, block_type)

    def resolve(self) -> None:
        """
        Resolve references by name in all registered objects.

        Replace block names by real block object.
        """
        for obj, attr, block_type in self._unresolved:
            blk = self._resolve_function(getattr(obj, attr))
            self._check_type(obj, attr, blk, block_type)
            setattr(obj, attr, blk)
        self._unresolved.clear()


class Circuit:
    """
    A container of all blocks and their interconnections.

    Not to be instantiated directly, rather call edzed.get_circuit().

    There exist only one circuit at a time and all created blocks
    belongs to it.

    In this implementation, circuit blocks can be added, but not removed.
    """

    def __init__(self):
        self._blocks: dict[str, block.Block] = {}
            # all blocks belonging to this circuit by name
        self._simtask: Optional[asyncio.Task] = None
            # the task running run_forever
        self._finalized: bool = False
            # no circuit modification after the initialization
        self._error: Optional[BaseException] = None
            # exception that terminated the simulator if any
        self.persistent_dict: Optional[MutableMapping[str, Any]] = None
            # persistent state data back-end
        self.persistent_ts: Optional[float] = None
            # timestamp of persistent data
        self.sblock_queue: asyncio.Queue
            # a Queue for notifying about changed SBlocks,
            # the queue will be created when simulation starts, because
            # it has a side effect of creating an event_loop if one
            # was not created yet. That may lead to errors in rare
            # scenarios with asyncio.new_event_loop().
        self._init_done: asyncio.Event
            # an Event for wait_init() synchronization
        self._resolver = _BlockResolver(self._validate_blk)
            # the name to block resolver
        self.resolve_name = self._resolver.register
            # provide access to the resolver
        self.debug: bool = False
            # enable debug messages

    def log_debug(self, *args, **kwargs) -> None:
        """Log a debug message if enabled."""
        if self.debug:
            _logger.debug(*args, **kwargs)

    def is_current_task(self) -> bool:
        """Return True if the simulator task is the current asyncio task."""
        if self._simtask is None:
            return False
        try:
            return self._simtask == asyncio.current_task()
        except Exception:
            # for instance RuntimeError('no running event loop')
            return False

    def is_finalized(self) -> bool:
        """Return True only if finalize() was called."""
        return self._finalized

    def is_ready(self) -> bool:
        """Return True only if ready to accept external events."""
        return self._simtask is not None and self._error is None

    @property
    def error(self) -> Optional[BaseException]:
        return self._error

    async def _check_started(self) -> None:
        """Raise if the simulation was not started yet."""
        if self._simtask is None:
            # just created?
            await asyncio.sleep(0)
            if self._simtask is None:
                raise EdzedInvalidState("The simulation task was not started")

    async def wait_init(self) -> None:
        """Wait until a running circuit is fully initialized."""
        await self._check_started()
        assert self._simtask is not None    # mypy type narrowing
        await asyncio.wait(
            [asyncio.create_task(self._init_done.wait()), self._simtask],
            return_when=asyncio.FIRST_COMPLETED)
        if self._simtask.done():
            if self._simtask.cancelled():
                msg = "The simulation task is finished"
            else:
                # normal simtask exit is not possible
                msg = f"The simulation task failed with error: {self._simtask.exception()}"
            raise EdzedInvalidState(msg)

    def check_not_finalized(self) -> None:
        """Raise an error if the circuit has been finalized."""
        if self._error:
            # there is an even bigger problem
            raise EdzedInvalidState("The circuit was shut down")
        if self._finalized:
            raise EdzedInvalidState("No changes allowed in a finalized circuit")

    def set_persistent_data(self, persistent_dict: Optional[MutableMapping[str, Any]]) -> None:
        """Setup the persistent state data storage."""
        self.check_not_finalized()
        self.persistent_dict = persistent_dict

    def addblock(self, blk: block.Block) -> None:
        """
        Add a circuit block. Not an ad blocker :-)

        Application code does not call this method, because
        blocks register themselves automatically when created.
        """
        self.check_not_finalized()
        if not isinstance(blk, block.Block):
            raise TypeError(f"Expected a Block object, got {blk!r}")
        if blk.name in self._blocks:
            raise ValueError(f"Duplicate block name {blk.name}")
        self._blocks[blk.name] = blk

    def getblocks(self, btype: Optional[type[block.Block|block.Addon]] = None) -> Iterator:
        """Return an iterator of all blocks or btype blocks only."""
        allblocks = self._blocks.values()
        if btype is None:
            return iter(allblocks)  # iter() only for type consistency, not really needed
        return (blk for blk in allblocks if isinstance(blk, btype))

    def findblock(self, name: str) -> block.Block:
        """Get block by name. Raise a KeyError when not found."""
        try:
            return self._blocks[name]
        except KeyError:
            raise KeyError(f"Block {name!r} not found") from None

    def set_debug(
            self,
            value: bool,
            *args: str|block.Block|type[block.Block|block.Addon]
            ) -> int:
        """
        Set debug flag to given value (True/False) for selected blocks.
        """
        todo = set()
        for arg in args:
            if isinstance(arg, str):
                if any(ch in arg for ch in '*?['):
                    for blk in self.getblocks():
                        if fnmatch.fnmatchcase(blk.name, arg):
                            todo.add(blk)
                else:
                    todo.add(self.findblock(arg))
            elif isinstance(arg, block.Block):
                todo.add(arg)
            # issubclass works only for classes
            elif isinstance(arg, type) and issubclass(arg, block.Block):
                for blk in self.getblocks(arg):
                    todo.add(blk)
            else:
                raise TypeError(f"Expected block name, Block object or subclass, got {arg!r}")
        value = bool(value)
        for blk in todo:
            blk.debug = value
        return len(todo)

    def _check_persistent_data(self) -> None:
        """Check persistent state related settings."""
        persistent_blocks = [
            blk for blk in self.getblocks(addons.AddonPersistence) if blk.persistent]
        if self.persistent_dict is None:
            if persistent_blocks:
                _logger.warning("No data storage, state persistence unavailable")
                for blk in persistent_blocks:
                    blk.persistent = False
            return

        try:
            self.persistent_ts = self.persistent_dict['edzed-stop-time']
            if not isinstance(self.persistent_ts, float):
                raise TypeError()
        except (KeyError, TypeError):
            self.persistent_ts = None
            _logger.warning(
                "The timestamp of persistent data is missing or invalid, "
                "state expiration will not be checked")
        else:
            if self.persistent_ts > time.time():
                _logger.error(
                    "The timestamp of persistent data is in the future, check the system time")

        # clear the unused items
        for key in self.persistent_dict.keys() - {blk.key for blk in persistent_blocks}:
            if key.startswith('edzed-'):
                continue
            _logger.info("Removing unused persistent state for '%s'", key)
            del self.persistent_dict[key]

    def _validate_blk(self, blk: Any) -> block.Block|block.Const:
        """
        Process a block specification. Return a valid block object.

        If the block is given by name, get the corresponding block.
        If a value is given, create a Const block from it.
        """
        if isinstance(blk, block.Const):
            return blk
        if isinstance(blk, str):
            # '_not_' + any 'NAME' except '_NAME'
            if blk.startswith('_') and blk not in self._blocks:
                # automatic blocks:
                if blk == '_ctrl':
                    return sblocks1.ControlBlock(
                        blk, comment="Simulation Control Block", _reserved=True)
                if blk.startswith('_not_') and blk[5:6] != '_':
                    return cblocks.Not(
                        blk, comment=f"Inverted output of {blk}", _reserved=True
                        ).connect(blk[5:])
            return self.findblock(blk)
        if not isinstance(blk, block.Block):
            return block.Const(blk)
        if blk not in self.getblocks():
            raise ValueError(f"{blk} is not in the current circuit")
        return blk

    def _finalize(self) -> None:
        """
        Complete the initialization of circuit block interconnections.

        For each block:
          - resolve temporary references by name
          - create inverter blocks for _not_name shortcuts
          - update connection data
        """
        def validate_output(iblk: block.CBlock, oblk: Any) -> block.Block|block.Const:
            """Validate oblk as a block to be connected to iblk."""
            try:
                return self._validate_blk(oblk)
            except Exception as err:
                add_note(err, f"failed connection: {oblk} --> {iblk}")
                raise

        for btype in (block.CBlock, cblocks.Not):
            # doing two passes because the first one may create new inverter blocks
            # some inverter blocks may be processed twice, that's acceptable
            for blk in list(self.getblocks(btype)):
                all_inputs: list[block.Block|block.Const] = []
                for iname, inp in blk.inputs.items():
                    if isinstance(inp, tuple):
                        newgroup = tuple(validate_output(blk, i) for i in inp)
                        all_inputs.extend(newgroup)
                        blk.inputs[iname] = newgroup
                    else:
                        newinput = validate_output(blk, inp)
                        all_inputs.append(newinput)
                        blk.inputs[iname] = newinput

                for inp in all_inputs:
                    if not isinstance(inp, block.Const):
                        blk.iconnections.add(inp)
                        self._blocks[inp.name].oconnections.add(blk)

    def finalize(self) -> None:
        """A wrapper for _finalize()."""
        if not self._finalized:
            self._finalize()
            self._finalized = True

    @staticmethod
    async def _run_tasks(
            jobname: str,
            btt_list: Iterable[tuple[block.Block, asyncio.Task, float]]
            ) -> None:
        """
        Run multiple tasks concurrently. Log errors.

        jobname is a description (e.g. "init" or "stop") used
        in messages. btt_list is a list of (block, task, timeout).
        It must not be empty.
        """
        assert btt_list
        errcnt = 0
        get_time = asyncio.get_running_loop().time
        start_time = get_time()
        for blk, task, timeout in sorted(btt_list, key=operator.itemgetter(2), reverse=True):
            # sorted from longest timeout
            if not task.done():
                try:
                    await asyncio.wait_for(task, timeout - get_time() + start_time)
                except asyncio.TimeoutError:
                    errcnt += 1
                    blk.log_warning(
                        "%s timeout, check timeout value (%.1f s)", jobname, timeout)
                except Exception:
                    # will be logged below
                    pass
            if not task.cancelled():
                err = task.exception()
                if err is not None:
                    errcnt += 1
                    blk.log_error("%s error: %s", jobname, err, exc_info=err)
        if errcnt:
            _logger.error("%d block %s error(s) suppressed", errcnt, jobname)

    async def _init_sblocks_async(self) -> None:
        """Initialize all sequential blocks, the async part."""
        start_tasks = [
            (
                blk,
                asyncio.create_task(
                    blk.init_async(), name=f"edzed: init_async for block {blk.name!r}"),
                blk.init_timeout
            ) for blk in self.getblocks(addons.AddonAsync)
            if not blk.is_initialized()
                and blk.has_method('init_async')
                and blk.init_timeout > 0.0]
        if start_tasks:
            self.log_debug("Initializing async sequential blocks")
            await self._run_tasks("async init", start_tasks)

    @staticmethod
    def init_sblock(blk: block.SBlock, full: bool) -> None:
        """
        Initialize a SBlock skipping any async code.

        The simulator is supposed to do either two calls with full=False
        or one call with full=True for each SBlock. The progress is
        registered in blk.init_steps_completed.

        On the first call:
            1.  call init_from_persistent_data - if applicable

        On the second call or if full=True
            2A. call init_regular
            2B. if not initialized, call init_from_value
                with the initdef value - if applicable

        After all initialization steps of init_sblock, a block:
            - MAY NOT be initialized, but
            - MUST be able to process events. An event is block's last
              chance to get its initialization. The simulation will
              fail unless all blocks are initialized properly.
        """
        steps = blk.init_steps_completed
        try:
            if steps == 0:
                if isinstance(blk, addons.AddonPersistence) and blk.persistent:
                    blk.init_from_persistent_data()
                    if blk.is_initialized():
                        blk.log_debug("initialized from saved state")
                blk.init_steps_completed = 1
            if steps == 1 or steps == 0 and full:
                blk.init_regular()
                if not blk.is_initialized() and blk.has_method('init_from_value') \
                        and blk.initdef is not block.UNDEF:
                    blk.init_from_value(blk.initdef)
                blk.init_steps_completed = 2
        except Exception as err:
            # add the block name
            add_note(err, f"block: {blk}, initialization error")
            raise

    def _init_sblocks_sync_1(self) -> None:
        """Initialize all sequential blocks, the non async part 1."""
        for blk in self.getblocks(block.SBlock):
            self.init_sblock(blk, full=False)

    def _init_sblocks_sync_2(self) -> None:
        """Initialize all sequential blocks, the non async part 2."""
        for blk in self.getblocks(block.SBlock):
            self.init_sblock(blk, full=False)
            # do not test yet, because the block might be still uninitialized
            # and waiting for an event that will be sent during another block's init
        for blk in self.getblocks(block.SBlock):
            if not blk.is_initialized():
                raise EdzedCircuitError(f"{blk}: not initialized")
        # save the internal states after initialization
        if self.persistent_dict is not None:
            for blk in self.getblocks(addons.AddonPersistence):
                blk.save_persistent_state()
        # clear the queue, the simulator knows that it needs to evaluate everything
        queue = self.sblock_queue
        while not queue.empty():
            queue.get_nowait()

    # AbstractSet does not define .difference and .intersection
    async def _stop_sblocks(self, blocks: set[block.SBlock]|frozenset[block.SBlock]):
        """
        Stop sequential blocks. Wait until all blocks are stopped.

        Stop blocks with async cleanup first, then all remaining blocks.

        Suppress errors.
        """
        async_blocks = {
            blk for blk in blocks.intersection(self.getblocks(addons.AddonAsync))
            if blk.has_method('stop_async') and blk.stop_timeout > 0.0}
        sync_blocks = blocks.difference(async_blocks)

        # 1. async blocks
        if async_blocks:
            for blk in async_blocks:
                try:
                    blk.stop()
                except Exception:
                    _logger.error("%s: ignored error in stop()", blk, exc_info=True)

            await asyncio.sleep(0)
            wait_tasks = [
                (
                    blk,
                    asyncio.create_task(
                        blk.stop_async(), name=f"edzed: stop_async for block {blk.name!r}"),
                    blk.stop_timeout
                ) for blk in async_blocks]
            self.log_debug("Waiting for async cleanup")
            await self._run_tasks("stop", wait_tasks)

        # 2. sync blocks
        for blk in sync_blocks:
            try:
                blk.stop()
            except Exception:
                _logger.error("%s: ignored error in stop()", blk, exc_info=True)

    async def _simulate(self) -> NoReturn:
        """
        Simulate the circuit until cancelled.

        On first run evaluate all blocks with inputs (i.e. CBlocks).
        Then evaluate all blocks directly connected to:
            - changed sequential blocks
            - changed evaluated blocks
        """
        def select_blk(block_set: Set[block.CBlock]) -> block.CBlock:
            """
            Return one block from the given set with minimum number of
            its inputs connected to another block within the set.
            The set must not be empty.
            """
            min_idep = min_blk = None
            for blk in block_set:
                idep = sum(1 for inp in blk.iconnections if inp in block_set)
                if idep == 0:
                    # 0 is the absolute minimum, no need to search further
                    return blk
                if min_idep is None or idep < min_idep:
                    min_idep = idep
                    min_blk = blk
            assert min_blk is not None  # because block_set must not be empty
            return min_blk

        eval_limit = _MAX_EVALS_PER_BLOCK * len(self._blocks)
        eval_set = set(self.getblocks(block.CBlock))
        eval_cnt = 0
        queue = self.sblock_queue
        while True:
            if not eval_set and queue.empty():
                self.log_debug("%d block(s) evaluated, pausing", eval_cnt)
                blk = await queue.get()
                self.log_debug("output change in %s, resuming", blk)
                eval_cnt = 0
                eval_set |= blk.oconnections
            while not queue.empty():
                blk = queue.get_nowait()
                eval_set |= blk.oconnections
            if not eval_set:
                continue
            eval_cnt += 1
            if eval_cnt > eval_limit:
                raise EdzedCircuitError(
                    "Circuit instability detected (too many block evaluations)")
            if len(eval_set) == 1:
                blk = eval_set.pop()
            else:
                blk = select_blk(eval_set)
                eval_set.discard(blk)
            try:
                changed = blk.eval_block()
            except Exception as err:
                # add the block name
                add_note(err, f"block: {blk}, output evaluation error")
                raise
            if changed:
                eval_set |= blk.oconnections

    async def run_forever(self) -> NoReturn:
        """
        Run the circuit simulation until the coroutine is cancelled.

        run_forever() never exits normally without an exception. It must
        be cancelled to stop the simulation.

        Please note that the cleanup could take some time - up to the
        max of SBlocks' stop_timeout values.

        When run_forever terminates, it cannot be invoked again.

        See also the ready attribute and wait_init().
        """
        if self._simtask is not None:
            if self._simtask.done():
                msg = "Cannot restart a finished simulation."
            else:
                msg = "The simulator is already running."
            raise EdzedInvalidState(msg)
        self._simtask = asyncio.current_task()
        started_blocks = set()
        start_ok = False
        try:
            if self._error is not None:
                raise self._error       # stop before start
            if not self._blocks:
                raise EdzedCircuitError("The circuit is empty")

            self.log_debug("Initializing the circuit")
            self.sblock_queue = asyncio.Queue()
            self._init_done = asyncio.Event()
            self._check_persistent_data()
            self._resolver.resolve()
            self.finalize()

            self.log_debug("Setting up circuit blocks")
            for blk in self.getblocks():
                blk.start()
                started_blocks.add(blk)
            # return control to the loop in order to run any tasks created by start()
            await asyncio.sleep(0)
            start_ok = True
            self.log_debug("Initializing sequential blocks")
            self._init_sblocks_sync_1()
            await self._init_sblocks_async()
            self._init_sblocks_sync_2()

            if self._error is None:
                self.log_debug("Starting simulation")
                self._init_done.set()
                await self._simulate()
                # not reached
                assert False
        # CancelledError is derived from BaseException, not Exception
        except (Exception, asyncio.CancelledError) as err:
            if self._error is None:
                self._error = err

        # Normally when a function from the try-except clause above calls abort(), the
        # abort() sets the self._error and cancels the task. The exception clause then
        # catches the cancellation.
        # But when a function calls abort() and also raises, the exception clause
        # catches the exception and the cancellation is left pending. For this
        # special edge case we must add a second except clause below.
        try:
            await asyncio.sleep(0)  # allow delivery of pending CancelledError if any
        except asyncio.CancelledError:
            pass

        if isinstance(self._error, asyncio.CancelledError):
            _logger.info("Normal circuit simulation stop")
        else:
            _logger.critical(
                "Fatal circuit simulation error: %s", self._error, exc_info=self._error)

        if started_blocks:
            # if blocks were started (at least some of them), they must be stopped;
            # save the state first, because stop may invalidate the state information
            if start_ok and self.persistent_dict is not None:
                for blk in started_blocks.intersection(self.getblocks(addons.AddonPersistence)):
                    blk.save_persistent_state()
                self.persistent_dict['edzed-stop-time'] = time.time()
            await self._stop_sblocks(started_blocks)
        assert self._error is not None  # mypy type narrowing
        raise self._error

    def abort(self, exc: BaseException) -> None:
        """
        Abort the circuit simulation due to an exception.

        abort() is necessary only when an ordinary exception wouldn't
        be propagated to the simulator.

        The first error stops the simulation, that's why abort()
        delivers the exception only if the simulation hasn't received
        another exception already.

        abort() may be called even before the simulation begins.
        The start will then fail.
        """
        if self._error is not None:
            if exc != self._error and exc != self._error.__cause__ \
                    and not isinstance(exc, asyncio.CancelledError):
                _logger.warning("ignoring subsequent abort(%r)", exc)
            return
        if not isinstance(exc, BaseException):
            # one more reason to abort
            exc = TypeError(f'abort(): expected an exception, got {exc!r}')
        _logger.warning("abort(%r)", exc)
        self._error = exc
        if self._simtask is not None and not self._simtask.done():
            self._simtask.cancel()

    async def shutdown(self) -> None:
        """
        Stop the simulation and wait until it finishes.
        """
        await self._check_started()
        assert self._simtask is not None    # mypy type narrowing
        if self.is_current_task():
            raise EdzedInvalidState("Cannot await the simulator task from the simulator task.")
        self.abort(asyncio.CancelledError('shutdown'))
        try:
            await self._simtask
        except asyncio.CancelledError:
            pass


def _install_sigterm_handler() -> None:
    old_handler = signal.getsignal(signal.SIGTERM)
    def handler(signum, frame):
        # using the _threadsafe variant because we need also to wake up the event loop
        call_soon = asyncio.get_running_loop().call_soon_threadsafe
        call_soon(_logger.warning, "SIGTERM caught")
        call_soon(get_circuit().abort, asyncio.CancelledError("SIGTERM caught"))
        if callable(old_handler):
            old_handler(signum, frame)
    signal.signal(signal.SIGTERM, handler)


async def run(*coroutines: Coroutine, catch_sigterm: bool = True) -> None:
    """
    Run Circuit.run_forever() and supporting coroutines as tasks.

    If any of the tasks exits, cancel all remaining tasks.

    Return None if all tasks exited normally or were cancelled.
    If the simulator raises, re-raise. If any of the supporting
    tasks raises, raise RuntimeError.
    """
    if catch_sigterm:
        # the signal handlers are chained, we will left the handler installed
        _install_sigterm_handler()

    circuit = get_circuit()
    if not coroutines:
        # do not create a needless task for this trivial case
        try:
            await circuit.run_forever()
        except asyncio.CancelledError:
            pass
        return

    simtask = asyncio.create_task(circuit.run_forever(), name="edzed: simulation task")
    tasks =  [
        asyncio.create_task(coro, name=f"edzed: supporting task #{i}")
        for i, coro in enumerate(coroutines, start=1)
        ] + [simtask]
    try:
        await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    except asyncio.CancelledError:
        pass
    for task in tasks:
        if not task.done():
            if task is simtask:
                # a direct simtask cancel could abort the cleanup
                circuit.abort(asyncio.CancelledError("shutdown"))
            else:
                task.cancel()
    sim_error = error_data = None
    for tnum, task in enumerate(tasks):
        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception as err:
            _logger.error("Error in task %s: %r", task, err)
            if task is simtask:
                sim_error = err
            else:
                error_data = (err, coroutines[tnum].__name__)

    if sim_error is not None:
        raise sim_error
    if error_data is not None:
        error, name = error_data
        raise RuntimeError(f"Supporting coroutine '{name}' in edzed.run() failed") from error



"""
Sequential blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any, NoReturn, Optional

from .. import addons
from .. import block
from .. import utils
from ..exceptions import EdzedCircuitError


__all__ = ['ControlBlock', 'Counter', 'Repeat', 'ValuePoll']


class ControlBlock(block.SBlock):
    """
    Simulator control block.
    """

    def _event_shutdown(
            self, *,
            source: str = '<no-source-data>',
            **_data) -> None:
        exc = asyncio.CancelledError(f"{self}: shutdown requested by '{source}'")
        self.circuit.abort(exc)

    def _event_abort(
            self, *,
            source: str = '<no-source-data>',
            error: BaseException|str = '<no-error-data>',
            **_data) -> None:
        exc = EdzedCircuitError(f"{self}: error reported by '{source}': {error!r}")
        if isinstance(error, Exception):
            exc.__cause__ = error
        self.circuit.abort(exc)

    def init_regular(self) -> None:
        self.set_output(None)


class Counter(addons.AddonPersistence, block.SBlock):
    """
    Counter. If modulo is set to a number M, count modulo M.
    """

    def __init__(
            self, *args,
            modulo: Optional[int|float] = None,
            initdef: int|float = 0,
            **kwargs):
        """
        Set the optional modulo and the initial counter value (default = 0).
        """
        if modulo == 0:
            raise ValueError("modulo must not be zero")
        self._mod = modulo
        super().__init__(*args, initdef=initdef, **kwargs)

    def _setmod(self, value: int|float) -> int|float:
        output = value if self._mod is None else value % self._mod
        self.set_output(output)
        return output

    def _event_inc(self, *, amount: int|float = 1, **_data) -> int|float:
        return self._setmod(self._output + amount)

    def _event_dec(self, *, amount: int|float = 1, **_data) -> int|float:
        return self._setmod(self._output - amount)

    def _event_put(self, *, value: int|float, **_data) -> int|float:
        return self._setmod(value)

    def _event_reset(self, **_data) -> int|float:
        return self._setmod(self.initdef)

    init_from_value = _setmod
    _restore_state = _setmod


class Repeat(addons.AddonMainTask, block.SBlock):
    """
    Periodically repeat the last received event.
    """

    def __init__(
            self, *args,
            dest: str|block.SBlock,
            etype: str|block.EventType = 'put',
            interval: int|float|str,
            count: Optional[int] = None,
            **kwargs):
        if isinstance(etype, block.EventCond):
            raise ValueError("An EventCond event cannot be repeated.")
        self._repeated_event = block.Event(dest, etype)
        self._interval = utils.time_period(interval)
        if self._interval is None or self._interval <= 0.0:
            raise ValueError("interval must be positive")
        if count is not None and count < 0:
            # count = 0 (no repeating) is accepted
            raise ValueError("argument 'count' must not be negative")
        self._queue: asyncio.Queue
        self._count = count
        self._warning_logged = False
        super().__init__(*args, **kwargs)

    def init_regular(self) -> None:
        self.set_output(0)

    async def _maintask(self) -> NoReturn:
        repeating = False
        while True:
            if repeating:
                try:
                    data = await asyncio.wait_for(self._queue.get(), self._interval)
                    repeat = 0
                except asyncio.TimeoutError:
                    repeat += 1
            else:
                # avoid wait_for() overhead when not repeating an event
                data = await self._queue.get()
                repeat = 0

            if repeat > 0:  # skip the original event
                self.set_output(repeat)
                self._repeated_event.send(self, **data, repeat=repeat)
            repeating = self._count is None or repeat < self._count

    def _event(self, etype: str|block.EventType, data) -> None:
        if etype != self._repeated_event.etype:
            if not self._warning_logged:
                self.log_warning(
                    f"Unexpected event type {etype} (this error is logged only once)")
                self._warning_logged = True
            return
        # send the original event synchronously in order
        # not to conceal a possible forbidden loop
        data['orig_source'] = data.get('source')
        self.set_output(0)
        self._repeated_event.send(self, **data, repeat=0)
        self._queue.put_nowait(data)

    def start(self) -> None:
        super().start()
        self._queue = asyncio.Queue()


class ValuePoll(addons.AddonMainTask, addons.AddonAsyncInit, block.SBlock):
    """
    A source of measured or computed values.
    """

    def __init__(self, *args, func: Callable[[], Any], interval: int|float|str, **kwargs):
        self._func = func
        self._interval = utils.time_period(interval)
        if self._interval is None or self._interval <= 0.0:
            raise ValueError("interval must be positive")
        super().__init__(*args, **kwargs)

    async def _maintask(self) -> NoReturn:
        """Data acquisition task: repeatedly obtain a value."""
        while True:
            value = self._func()
            if asyncio.iscoroutine(value):
                value = await value
            if value is not block.UNDEF:
                self.set_output(value)
            await asyncio.sleep(self._interval)

    def init_from_value(self, value: Any) -> None:
        self.set_output(value)



"""
Conversion routines for time spans using multiple units.

Example: "20h15m10" = 20 hours + 15 minutes + 10 seconds = 72910 seconds
"""

from __future__ import annotations

from typing import overload
import re

from .tconst import *   # pylint: disable=wildcard-import, unused-wildcard-import

__all__ = ['timestr', 'timestr_approx', 'convert', 'time_period']

# pylint: disable=invalid-name
def timestr(seconds: int|float, sep: str = '', prec: int = 3) -> str:
    """
    Return seconds as a string using d, h, m and s units.

    The individual parts are separated with the 'sep' string.

    Fractional seconds are formatted to 'prec' decimal places.
    """
    if seconds < 0:
        raise ValueError("Number of seconds cannot be negative")
    is_float = isinstance(seconds, float)
    if is_float:
        seconds = round(seconds, prec)
    d, s = divmod(seconds, SEC_PER_DAY)
    h, s = divmod(s, SEC_PER_HOUR)
    m, s = divmod(s, SEC_PER_MIN)
    parts = []
    if d:
        parts.append(f"{int(d)}d")
    if d or h:
        parts.append(f"{int(h)}h")
    parts.append(f"{int(m)}m")
    parts.append(f"{s:.{prec}f}s" if is_float else f"{s}s")
    return sep.join(parts)

def timestr_approx(seconds: int|float, sep: str = '') -> str:
    """
    Return possibly rounded seconds as a string using d, h, m, s units.

    The individual parts are separated with the 'sep' string.
    """
    #      ge  lt*   format
    #     ---  ---   ------
    #           1s   0.sss (float)
    #      1s  10s   S.ss  (float)
    #     10s   1m   S.s   (float)
    #           1m   S     (int)
    #      1m  10h   (H) M S
    #     10h  10d   (D) H M
    #     10d        D H
    #
    # (*) lt (less than) after rounding. We want to prevent e.g. 0.9998
    # to be rounded to 1 with three decimal places 1.000, but 1.002 to
    # 1 with two decimal places 1.00. We prefer consistency and always
    # have 0.xxx and 1.yy.

    if seconds < 0:
        raise ValueError("Number of seconds cannot be negative")
    omit_minutes = omit_seconds = False
    if isinstance(seconds, float):
        # the rounding issues are trickier than it might seem ...
        if seconds < 1.0:
            sprec = 3   # seconds' precision (decimal places)
            seconds = round(seconds, sprec) # may be now equal to 1.0, no elif in the next line
        if 1.0 <= seconds < 10.0:
            sprec = 2
            seconds = round(seconds, sprec) # may be now equal to 10.0
        if 10.0 <= seconds < 60.0:
            sprec = 1
            seconds = round(seconds, sprec) # may be now equal ... etc.
        if 60.0 <= seconds < 10*SEC_PER_HOUR:
            seconds = round(seconds)        # now an int => 'sprec' unused
    # int or float
    if 10*SEC_PER_HOUR <= seconds < 10*SEC_PER_DAY:
        omit_seconds = True
        seconds = SEC_PER_MIN * int(seconds / SEC_PER_MIN + 0.5)    # round to nearest minute
    if 10*SEC_PER_DAY <= seconds:
        omit_seconds = omit_minutes = True
        seconds = SEC_PER_HOUR * int(seconds / SEC_PER_HOUR + 0.5)  # round to nearest hour

    d, s = divmod(seconds, SEC_PER_DAY)
    h, s = divmod(s, SEC_PER_HOUR)
    if not omit_minutes:
        m, s = divmod(s, SEC_PER_MIN)
    parts = []
    if d:
        parts.append(f"{int(d)}d")
    if d or h:
        parts.append(f"{int(h)}h")
    if not omit_minutes and (d or h or m):
        parts.append(f"{int(m)}m")
    if not omit_seconds:
        # seconds = original value; s = 0 to 60 seconds
        parts.append(f"{s:.{sprec}f}s" if isinstance(seconds, float) else f"{s}s")
    return sep.join(parts)

_RE_TIME = r'(?:(\d+)\s*d)?\s*(?:(\d+)\s*h)?\s*(?:(\d+)\s*m)?\s*(?:([\d.]+)\s*s?)?'
_REGEXP_TIME = re.compile(_RE_TIME, re.ASCII | re.IGNORECASE)

def convert(tstr: str) -> float:
    """
    Convert string to number of seconds. Return float.

    Format:
        TIMESTR = [DAYS] [HOURS] [MINUTES] [SECONDS]
    where:
        DAYS = <int> "D"
        HOURS = <int> "H"
        MINUTES = <int> "M"
        SECONDS =  <int or float> ["S"]
    Notes:
        - whitespace around numbers and units is allowed
        - numbers do not have to be normalized, e.g. "72h" is OK
        - unit symbols D, H, M, S may be entered in upper or lower case
        - negative values are not allowed
        - float values with exponents are not supported
    """
    match = _REGEXP_TIME.fullmatch(tstr.strip())
    if match is None:
        raise ValueError("Invalid time representation")
    d, h, m, s = match.groups()
    result = 0.0
    if d:
        result += int(d) * SEC_PER_DAY
    if h:
        result += int(h) * SEC_PER_HOUR
    if m:
        result += int(m) * SEC_PER_MIN
    if s:
        result += float(s)  # may raise ValueError, the regexp is not strict enough
    return result

@overload
def time_period(period: None) -> None:
    ...
@overload
def time_period(period: int|float|str) -> float:
    ...
def time_period(period):
    """Convenience wrapper for convert()."""
    if period is None:
        return None
    if isinstance(period, int):
        period = float(period)
    if isinstance(period, float):
        return max(0.0, period)
    if isinstance(period, str):
        return convert(period)
    raise TypeError(f"Invalid type for time period specification: {period!r}")



"""
A simple interactive utility for learning and evaluating edzed.

WARNING: Do not use for anything else. Contains "eval <user input>".

Usage:
    1. append following lines to a python file with an edzed
    circuit definition (e.g. circuit1.py):
    --- begin ---
    if __name__ == '__main__':
        import asyncio  # if not imported at the top
        import edzed.demo
        asyncio.run(edzed.run(edzed.demo.cli_repl())
    --- end ---
    2. run the demo:
        python3 circuit1.py

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from __future__ import annotations

import ast
import asyncio
import collections
from collections.abc import Callable
from dataclasses import dataclass
import logging
import signal
import sys
from typing import Any, Optional

from . import simulator
from . import block


__all__ = ['cli_repl', 'run_demo']

LIMIT = 4096    # StreamReader buffer limit
HISTSIZE = 20   # history list size

HELP = """\
Control commands:
    h[elp] or ?                 -- show this help
    exit
    eval <python_expression>
Circuit evaluation commands:
  Debug messages:
    a[debug] 1|0                -- all blocks' debug messages on|off
    b[debug] <blockname> 1|0    -- block's debug messages on|off
    c[debug] 1|0                -- circuit simulator's debug messages on|off
  Events:
    e[vent] <blockname> <type> [{'name':value, ...}]
                                -- send event with optional data
    p[ut] <blockname> <value>   -- send 'put' event
  Info:
    l[ist]                      -- list all blocks
    i[nfo] <blockname>          -- print block's properties
    s[how] <blockname>          -- print current state and output
Command history:
    !?                          -- print history
    !N                          -- repeat command N (integer)
    !-N                         -- repeat command current minus N
    !!                          -- repeat last command (same as !-1)
"""

def _check_01(value: str) -> bool:
    if value == '0':
        return False
    if value == '1':
        return True
    raise ValueError("Argument must be 0 (debug off) or 1 (debug on)")


def _cmd_adebug(value: str) -> None:
    bvalue = _check_01(value)
    circuit = simulator.get_circuit()
    circuit.set_debug(bvalue, *circuit.getblocks())
    print(f"all blocks: debug {'on' if bvalue else 'off'}")


def _cmd_bdebug(blk: block.Block, value: str) -> None:
    bvalue = _check_01(value)
    blk.debug = bvalue
    print(f"{blk}: debug {'on' if bvalue else 'off'}")


def _cmd_cdebug(value: str) -> None:
    bvalue = _check_01(value)
    circuit = simulator.get_circuit()
    circuit.debug = bvalue
    print(f"circuit simulator: debug {'on' if bvalue else 'off'}")


def _cmd_eval(expr: str) -> None:
    result = eval(expr)         # pylint: disable=eval-used
    print(f"result: {result}")


def _cmd_event(blk: block.SBlock, etype: str, data: Optional[str] = None) -> None:
    datadict = {} if data is None else ast.literal_eval(data)
    if not isinstance(datadict, dict):
        raise TypeError(
            f"Invalid event data: {datadict!r}. "
            "Expected is a dict literal: {{'name':value, ...}}")
    retval = blk.event(etype, **datadict)
    print(f"event() returned: {retval}")
    _cmd_show(blk)


def _cmd_help() -> None:
    print(HELP)


def _cmd_info(blk: block.Block) -> None:
    for name, value in sorted(blk.get_conf().items()):
        print(f"{name}: {value}")


def _cmd_list() -> None:
    circuit = simulator.get_circuit()
    for name, btype in sorted((blk.name, type(blk).__name__) for blk in circuit.getblocks()):
        print(f"{name}   ({btype})")


def _cmd_put(blk: block.SBlock, value: Any) -> None:
    retval = blk.put(ast.literal_eval(value))
    print(f"put() returned: {retval}")
    _cmd_show(blk)


def _cmd_show(blk) -> None:
    output = blk.output
    state = blk.get_state()
    if state == output:
        print(f"output and state: {output!r}")
    else:
        print(f"output: {output!r}")
        print(f"state: {state!r}")


def _complete(cmd: str) -> str:
    if cmd == '?':
        return 'help'
    for fullcmd in _PARSE:
        # no abbreviation for eval
        if fullcmd.startswith(cmd) and not 'eval' == fullcmd != cmd:
            return fullcmd
    raise ValueError(f"Unknown command: {cmd}, try: help")


@dataclass(frozen=True)
class CmdInfo:
    func: Callable
    blk: bool = True    # first arg is block name
    args: int = 0       # number of mandatory arguments (excl. the block name if blk is set)
    optargs: int = 0    # number of optional arguments

_PARSE = {
    'adebug': CmdInfo(func=_cmd_adebug, blk=False, args=1),
    'bdebug': CmdInfo(func=_cmd_bdebug, args=1),
    'cdebug': CmdInfo(func=_cmd_cdebug, blk=False, args=1),
    'eval': CmdInfo(func=_cmd_eval, blk=False, args=1),
    'event': CmdInfo(func=_cmd_event, args=1, optargs=1),
    # exit is handled in the loop
    'help': CmdInfo(func=_cmd_help, blk=False),
    'info': CmdInfo(func=_cmd_info),
    'list': CmdInfo(func=_cmd_list, blk=False),
    'put': CmdInfo(func=_cmd_put, args=1),
    'show': CmdInfo(func=_cmd_show),
}

_HFORMAT = " cmd {:2d}> {}"

async def _cli_repl() -> None:
    """
    Edzed demo CLI REPL.

    CLI = command line interface; REPL = read-evaluate-print loop.
    """
    loop = asyncio.get_running_loop()
    rstream = asyncio.StreamReader(limit=LIMIT, loop=loop)
    protocol = asyncio.StreamReaderProtocol(rstream, loop=loop)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)

    history: collections.deque[tuple[str, Callable, list]] = collections.deque(maxlen=HISTSIZE)
    cmdnum = 1
    circuit = simulator.get_circuit()
    print("Type 'help' to get a summary of available commands.")
    print("Type 'a 1' to enable debug messages in all blocks.")
    while True:
        await asyncio.sleep(0)
        print(f'--- edzed {cmdnum}> ', end='', flush=True)
        line = (await rstream.readline()).decode()
        if not line:
            print('received EOF')
            break
        line = line.strip()
        if not line:
            continue
        cmd, *args = line.split(None, 1)
        if cmd == 'exit':
            # arguments ignored
            break
        try:
            if cmd and cmd[0] == '!':
                # history
                if args:
                    raise TypeError("history: syntax error")
                cmd = cmd[1:]
                hlen = len(history)
                if cmd == '?':
                    for hnum, line_func_args in enumerate(history, start=cmdnum-hlen):
                        print(_HFORMAT.format(hnum, line_func_args[0]))
                    continue
                if cmd == '!':
                    hnum = -1
                else:
                    try:
                        hnum = int(cmd)
                    except ValueError:
                        raise ValueError("history !N: invalid command number N") from None
                if hnum < 0:
                    hnum += cmdnum
                if not 1 <= hnum < cmdnum:
                    raise LookupError(f"history: command {hnum} does not exist")
                idx = hnum + hlen - cmdnum
                # beware: no IndexError for small negative indices
                if not 0 <= idx < hlen:
                    raise LookupError(f"history: command {hnum} not in memory")
                line, func, args = history[hnum + hlen - cmdnum]
                print(_HFORMAT.format(hnum, line))
            else:
                # command
                cmd = _complete(cmd)
                cmdinfo = _PARSE[cmd]
                minargs = cmdinfo.args + (1 if cmdinfo.blk else 0)
                maxargs = minargs + cmdinfo.optargs
                if args != []:
                    args = args[0].split(None, maxargs - 1)
                if not minargs <= len(args) <= maxargs:
                    if minargs == maxargs:
                        raise TypeError(f"{cmd} command takes {minargs} argument(s)")
                    raise TypeError(f"{cmd} command takes {minargs} to {maxargs} arguments")
                if cmdinfo.blk:
                    args[0] = circuit.findblock(args[0])
                func = cmdinfo.func
            # execute
            func(*args)
            history.append((line, func, args))
            cmdnum += 1
        except Exception as err:
            print(f"ERROR: {err}")


async def cli_repl(setup_logging: bool = True) -> None:
    """
    A wrapper preparing the run environment for _cli_repl().

    Set up:
    - SIGINT handler for better UX.
    - logging configuration that displays debug messages

    Refer to _cli_repl().
    """
    if setup_logging:
        logging.basicConfig(level=logging.DEBUG)

    task = asyncio.current_task()
    def sigint_handler(_signum, _frame):
        # using the _threadsafe variant because we need also to wake up the event loop
        call_soon = asyncio.get_running_loop().call_soon_threadsafe
        call_soon(print, " -- Interrupt signal received, exiting the edzed demo")
        call_soon(task.cancel)
        # NOT calling the Python's default interrupt handler
    saved_sigint_handler = signal.signal(signal.SIGINT, sigint_handler)
    try:
        await _cli_repl()
    finally:
        # revert to the original state, because we broke the SIGINT handlers chain
        signal.signal(signal.SIGINT, saved_sigint_handler)



"""
Event filters.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from __future__ import annotations

from collections.abc import Callable, Mapping, MutableMapping
import logging
import types
from typing import Any, cast

from .. import block
from .. import simulator


__all__ = ['not_from_undef', 'Edge', 'Delta', 'DataEdit', 'IfOutput', 'IfNotIitialized']

_logger = logging.getLogger(__package__)


def not_from_undef(data: Mapping) -> bool:
    """Filter out the initial change from UNDEF to the first real value."""
    return data.get('previous', block.UNDEF) is not block.UNDEF


class Edge:
    """
    Event filter for logical values.
    """

    def __init__(
            self,
            rise: bool = False,
            fall: bool = False,
            u_rise: bool|None = None,
            u_fall: bool = False):
        self._rise = bool(rise)
        self._fall = bool(fall)
        self._urise = bool(u_rise) if u_rise is not None else self._rise
        self._ufall = bool(u_fall)
        if not (rise or fall or u_rise or u_fall):
            _logger.warning(
                "%s: all events will be filtered out!",
                type(self).__name__)

    def __call__(self, data: Mapping) -> bool:
        value = data['value']
        previous = data['previous']
        if previous is block.UNDEF:
            if self._urise if value else self._ufall:
                return True
        else:
            previous = bool(previous)
            if (not previous and self._rise) if value else (previous and self._fall):
                return True
        return False


class Delta:
    """
    Event filter for numeric values.
    """

    def __init__(self, delta: int|float):
        self._delta = delta
        self._last = block.UNDEF

    def __call__(self, data: Mapping) -> bool:
        value = data['value']
        if self._last is block.UNDEF or abs(self._last - value) >= self._delta:
            self._last = value
            return True
        return False


class IfOutput:
    """
    Enable/disable events depending on block's output.
    """

    def __init__(self, control_block: str|block.Block):
        self._ctrl_blk = control_block
        simulator.get_circuit().resolve_name(self, '_ctrl_blk')

    def __call__(self, data: Mapping) -> Mapping|None:
        return data if cast(block.Block, self._ctrl_blk).output else None


class IfNotIitialized:
    """
    Enable/disable events depending on block's init state.
    """

    def __init__(self, control_block: str|block.SBlock):
        self._ctrl_blk = control_block
        simulator.get_circuit().resolve_name(self, '_ctrl_blk', block_type=block.SBlock)

    def __call__(self, data: Mapping) -> Mapping|None:
        return None if cast(block.SBlock, self._ctrl_blk).is_initialized() else data


class dualmethod(classmethod):
    """
    Dual (class/instance) method decorator.

    When the decorated method is called as a class method,
    create an instance on the fly.

    When called as an instance method, proceed normally,
    i.e. as if not decorated.
    """

    def __get__(self, instance, cls):
        if instance is None:
            instance = cls()
        return self.__func__.__get__(instance, cls)     # pylint: disable=no-member


class DataEdit:
    """
    Modify the event data.

    Methods may be chained.
    """

    def __init__(self):
        self._editlist = []

    # @dualmethod confuses pylint a little
    # pylint: disable=bad-classmethod-argument, no-member
    @dualmethod
    def add(self, **kwargs) -> DataEdit:
        """Add key=value pairs. Existing values will be overwritten."""
        self._editlist.append(lambda data: {**data, **kwargs})
        return self

    @dualmethod
    def add_output(self, key: str, source: str|block.Block) -> DataEdit:
        """Add key=block's output. Existing value will be overwritten."""
        # cannot store the 'source' as an instance attribute, because
        # next 'add_output' call would overwrite it. In order to prevent
        # that a separate container must be created each time.
        src = types.SimpleNamespace(block=source)
        simulator.get_circuit().resolve_name(src, 'block')
        self._editlist.append(lambda data: {**data, key: src.block.output})
        return self

    @dualmethod
    def copy(self, src: str, dst: str) -> DataEdit:
        """Copy data[src] to data[dst]."""
        def _edit(data: MutableMapping) -> MutableMapping:
            data[dst] = data[src]
            return data
        self._editlist.append(_edit)
        return self

    @dualmethod
    def delete(self, *args: str) -> DataEdit:
        """Delete listed keys. Non-existing keys are ignored."""
        def _edit(data: MutableMapping) -> MutableMapping:
            for key in args:
                data.pop(key, None)
            return data
        self._editlist.append(_edit)
        return self

    DELETE = object()
    REJECT = object()

    @dualmethod
    def modify(self, key: str, func: Callable[[Any], Any]) -> DataEdit:
        """Apply the func to a value identified by key."""
        def _edit(data: MutableMapping) -> MutableMapping|None:
            current = data[key]
            replacement = func(current)
            if replacement is self.REJECT:
                return None
            if replacement is self.DELETE:
                del data[key]
            else:
                data[key] = replacement
            return data
        self._editlist.append(_edit)
        return self

    @dualmethod
    def permit(self, *args) -> DataEdit:
        """Delete all but listed keys."""
        def _edit(data: MutableMapping) -> MutableMapping:
            for key in list(data):
                if key not in args:
                    del data[key]
            return data
        self._editlist.append(_edit)
        return self

    @dualmethod
    def rename(self, src, dst) -> DataEdit:
        """Rename key: data[src] -> data[dst]."""
        def _edit(data: MutableMapping) -> MutableMapping:
            data[dst] = data[src]
            del data[src]
            return data
        self._editlist.append(_edit)
        return self

    @dualmethod
    def setdefault(self, **kwargs) -> DataEdit:
        """Add key=value pairs only if key is missing."""
        self._editlist.append(lambda data: {**kwargs, **data})
        return self

    def __call__(self, data: MutableMapping) -> MutableMapping|None:
        for func in self._editlist:
            data = func(data)
            if not isinstance(data, MutableMapping):
                break
        return data
